<?php
/**
 * Giscuit
 *
 * LICENSE
 *
 * This source file is subject to the Giscuit license that is bundled
 * with this application in the file LICENSE.txt.
 *
 * @version    $Id: Map.php 3480 2013-06-27 08:48:05Z pavel.iacovlev $
 * @category   Giscuit
 * @package    Giscuit_Map
 * @copyright  Copyright (c) 2012 VEC
 */
/**
 * @see Giscuit_Map_Abstract
 */
require_once 'Giscuit/Map/Abstract.php';

/**
 * Class for abstract mapserver object
 *
 * @category   Giscuit
 * @package    Giscuit_Map
 * @copyright  Copyright (c) 2012 VEC
 */
class Giscuit_Map_Map extends Giscuit_Map_Abstract
{

    /**
     * Map scale
     *
     * @var array
     */
    protected $_scale;

    /**
     * Map extent
     *
     * @var array
     */
    protected $_extent;

    /**
     * Pixel size
     *
     * @var array
     */
    protected $_pixelSize;

    /**
     * Resolution
     * 
     * @var integer
     */
    CONST RESOLUTION = 72;

    /**
     * Map extent
     *
     * @var array
     */
    protected $_size;

    /**
     * Used for map printing, edit layer will be visibile  
     *
     * @var bool
     */
    public $addEditLayer = false;

    /**
     * Contains data of the map legend
     * 
     * @var array
     */
    public $legendArr = array();

    /**
     * Is $legendArr being populated with map legend data
     * 
     * @var bool
     */
    public $legend = false;

    /**
     * Load user defined settings
     * 
     * @var bool
     */
    public $withLayerSettings;
    
    /**
     * Maximum image width
     * 
     * @var integer
     */
    public $maxWidth = 1920;
    
    /**
     * Maximum image height
     * 
     * @var integer
     */
    public $maxHeight = 1200;
    
    /**
     * Dump all of the layer attributes
     * 
     * @var bool
     */
    public $dumpAllAttributes = false;
    
    /**
     * Class constructor
     *
     * @param string $mapFile Full path to the mapserver configuration file
     * @return void
     */
    public function __construct ($mapFile)
    {
        $this->_obj = ms_newMapObj($mapFile);
        
        if(GT_TESTING == true) {
            $projLibPath = GT_MS_PATH . DIR_SEP . 'proj' . DIR_SEP . 'nad';
            if(file_exists($projLibPath)) {
                $this->getMsObj()->setconfigoption('PROJ_LIB', $projLibPath);
            }
        }
        
        $this->setExtent($this->getExtent());
        $this->setSize($this->getSize());
        $this->setPixelSize();
    }

    /**
     * Set default map parameters
     * 
     * @return void
     */
    public function setParamsFromConfig ()
    {
        $config = Zend_Registry::get('config');
        
        $this->set('units', $this->unitsToValue($config->map->units));
        $this->setProjection('init=epsg:' . $config->map->srid);
        $this->setFontSet(GT_CONFIG_PATH . DIR_SEP . 'fonts' . DIR_SEP . 'fonts.txt');
        $this->setSymbolSet(GT_CONFIG_PATH . DIR_SEP . 'symbols.sym');
        
        $this->web->set('imagepath', GT_APP_PATH . DIR_SEP . 'tmp');
        
        $this->set('shapepath', GT_RASTER_PATH);
    }

    /**
     * Get map scale
     * 
     * @return int The normalized scale value
     */
    public function getScale ()
    {
        $config = Zend_Registry::get('config');
        
        $srid = NULL;
        $scale = NULL;
        
        if(! empty($_GET['SRS'])) $srid = $_GET['SRS'];
        if(! empty($_GET['srs'])) $srid = $_GET['srs'];
        if(! empty($_GET['CRS'])) $srid = $_GET['CRS'];
        if(! empty($_GET['crs'])) $srid = $_GET['crs'];               
        
        if(! empty($srid)) {
            if(strtolower(substr($srid, 0, 4)) == 'epsg' && intval(substr($srid, 5)) != $config->map->srid) {
                $srid = intval(substr($srid, 5));
                
                //Commit
                $this->setProjection('init=epsg:' . $srid, MS_TRUE);
                $this->setExtent($this->getExtent());
                $this->preparequery();
                $scale = $this->scaledenom;
                
                //Rollback
                $this->setProjection('init=epsg:' . $config->map->srid, MS_TRUE);
                $this->setExtent($this->getExtent());
                $this->preparequery();
            }  
        }
        
        if(empty($scale)) {
            $this->preparequery();
            $scale = $this->scaledenom;
        }
        
        return $scale;
    }
    
    /**
     * Set pixel size
     * 
     * @return void
     */
    public function setPixelSize ()
    {
        if(! empty($this->_size)) {
            $extent = $this->getExtent();
            $this->_pixelSize['width'] = ($extent['max']['x'] - $extent['min']['x']) / $this->_size['width'];
            $this->_pixelSize['height'] = ($extent['max']['y'] - $extent['min']['y']) / $this->_size['height'];
        }
    }

    /**
     * Get pixel size
     * 
     * @return array
     */
    public function getPixelSize ()
    {
        return $this->_pixelSize;
    }

    /**
     * Set map extent
     * 
     * @return void
     */
    public function setExtent (array $extent = null)
    {
        $this->getMsObj()->setExtent($extent['min']['x'], $extent['min']['y'], $extent['max']['x'], $extent['max']['y']);
        $this->_extent = $extent;
    }

    /**
     * Get map extent
     * 
     * @return array
     */
    public function getExtent ()
    {
        $config = Zend_Registry::get('config');
        
        $params = array_change_key_case($_GET);
        
        //From object
        if (! empty($this->_extent)) {
            return $this->_extent;
        }
        
        if (empty($params['mapext']) && empty($params['bbox'])) {
            return $config->map->extent->toArray();
        }
        if (! empty($params['mapext'])) {
            $mapext = $params['mapext'];
            $mapext = explode(' ', $mapext);
        } else {
            $mapext = $params['bbox'];
            $mapext = explode(',', $mapext);
        }
        
        $extent = array();
        $extent['min']['x'] = floatval($mapext[0]);
        $extent['min']['y'] = floatval($mapext[1]);
        $extent['max']['x'] = floatval($mapext[2]);
        $extent['max']['y'] = floatval($mapext[3]);
        
        return $extent;
    }

    /**
     * Set map size
     * 
     * @return void
     */
    public function setSize ($size)
    {
        if(! empty($size)) {
            $this->getMsObj()->setSize($size['width'], $size['height']);
            $this->_size = $size;
        }
    }

    /**
     * Get map size
     * 
     * @return array
     */
    public function getSize ()
    {
        $config = Zend_Registry::get('config');
        
        if (! empty($this->_size)) {
            return $this->_size;
        }
        
        //Map size
        if (! empty($_GET['map_size'])) {
            $mapSize = $_GET['map_size'];
        } elseif (! empty($_GET['MAP_SIZE'])) {
            $mapSize = $_GET['MAP_SIZE'];
        }
        if (! empty($mapSize)) {
            $mapSize = explode(' ', $_GET['map_size']);
            if (count($mapSize) == 2) {
                $size['width'] = floatval($mapSize[0]);
                $size['height'] = floatval($mapSize[1]);
            }
        }
        
        //Width
        if(isset($_GET['width'])) {
            $size['width'] = $_GET['width'];
        } else if(isset($_GET['WIDTH'])) {
            $size['width'] = $_GET['WIDTH'];
        }
        
        //Height
        if(isset($_GET['height'])) {
            $size['height'] = $_GET['height'];
        } else if(isset($_GET['HEIGHT'])) {
            $size['height'] = $_GET['HEIGHT'];
        }
                
        //
        if(empty($size['width']) || empty($size['height'])) {
            return;
        }
        
        //FIXME:
        if(! isset($_GET['SERVICE'])) {
            $_GET['SERVICE'] = '';
        }
        if(! isset($_GET['service'])) {
            $_GET['service'] = '';
        }
        
        //
        if ($size['width'] > $this->maxWidth 
            && $size['height'] > $this->maxWidth
            && strtolower($_GET['SERVICE']) != 'wfs'
            && strtolower($_GET['service']) != 'wfs'
            ) {
                
            throw new Giscuit_Exception('Invalid image size');
        }
        
        $size['width'] = intval($size['width']);
        $size['height'] = intval($size['height']);
        
        return $size;
    }

    /**
     * Add layers defined in the layers table to the map object
     * 
     * @param $options <ul>Available options:
     * <li>withLayerSettings bool - Load user defined settings</li>
     * <li>addEditLayer bool - Used for map printing, edit layer will be visibile</li>
     * <li>watermark bool - Watermark content</li></ul>
     * @return void
     */
    public function addCurrentLayers (array $options = array())
    {
        $config = Zend_Registry::get('config');
        $translate = Zend_Registry::get('Zend_Translate');

        $this->addEditLayer = $options['addEditLayer'];
        
        if (empty($options['watermark'])) {
            $options['watermark'] = NULL;
        }
        
        if (empty($options['withLayerSettings'])) {
            $this->withLayerSettings = false;
        } else {
            $this->withLayerSettings = true;
        }
        
        //Layer settings
        $layerTable = new Giscuit_Db_Table_Layers();
        $layers = $layerTable->getItems($options);
        
        $outputFormat = NULL;
        //Add layers
        foreach ($layers as $settings) {
            $layer = $compareLayer = NULL;
            
            //Skip group
            if (empty($settings['class'])) {
                continue;
            }
            
            //Skip javascript layers
            if(is_subclass_of('Giscuit_Map_Layer_' . $settings['class'], 'Giscuit_Map_Layer_JavaScript') &&
                $settings['js_isbase'] == 't') {
                    
                $outputFormat = 'png';
                continue;
            }                    
            
            $layerName = 'Giscuit_Map_Layer_' . $settings['class'];
            $tableName = 'Giscuit_Db_Table_Layer_' . $settings['class'];
            
            $layer = new $layerName($this, $settings);
            
            //Add compare layer
            if(isset(Zend_Registry::get('settings')->historyId)) {
                $historyId = Zend_Registry::get('settings')->historyId;
            }
            if (! empty($historyId) &&
                @class_exists($tableName) 
                && @class_exists('Giscuit_Db_Table_Layer_History') 
                && is_subclass_of($tableName, 'Giscuit_Db_Table_Layer_History')) {
                    
                $compareLayer = new $layerName($this, array_merge(array(
                    'whereHistoryId' => $historyId
                ), $settings));
                
                $layer->set('opacity', Giscuit_Controller_Plugin_Editing::$layerOpacity);   
            }
            
        }
        
        if($outputFormat == NULL) {
            $this->selectOutputFormat('jpeg');
        } else {
            $this->selectOutputFormat('png');
            $this->outputformat->set('imagemode', MS_IMAGEMODE_RGBA);
            $this->outputformat->transparent = MS_TRUE;
        }
        
        $license = getLicense();
        if($license['properties']['type']['value'] == 0) {

        }
        
        //Watermark
        $auth = Zend_Auth::getInstance();
        if ($auth->hasIdentity()) {
            if ($config->map->watermarks == 't' &&
                $license['properties']['type']['value'] != 0) {
                    
                $layer = new Giscuit_Map_Layer_Copyright($this, array(
                    'class' => 'Copyright' , 
                    'watermark' => $options['watermark']
                ));
            }
            
            if ($license['properties']['type']['value'] == 0) {
                $expireDate = new Zend_Date($license['info']['FILE_EXPIRY']);
                $timeleft = $expireDate->get(GT_DATETIME_STYLE_PHP);
                
                $layer = new Giscuit_Map_Layer_Copyright($this, array(
                    'class' => 'Copyright' , 
                    'watermark' => $translate->_('trialExpiresAt') . "\n" . $timeleft
                ));
            }
            
        }
        
        //Mask
        $maskingPlugin = Zend_Controller_Front::getInstance()->getPlugin('Giscuit_Controller_Plugin_Masking');
        if ($maskingPlugin != NULL) {
            $isAllowed = Zend_Controller_Action_HelperBroker::getStaticHelper('isAllowed');
            
            if (! $isAllowed->direct('admin')) {
                $maskLayerClass = $maskingPlugin->layerClass;
                $maskLayerName = 'Giscuit_Map_Layer_' . $maskLayerClass;
                
                $layer = new $maskLayerName($this, array(
                    'class' => $maskLayerClass
                ));
            }
        }
        
        //Legend
        if ($this->legend == true) {
            foreach($this->children as $legendLayer) {
                foreach ($legendLayer->children as $legendClass) {
                    $legendClass->createLegend();
                }
            }   
        }
        
        return $layers;
    }

    /**
     * Convert bbox wkt representation into an array
     * 
     * @param string $geometry Geometry WKT string
     * @return array 
     */
    public static function bboxToExtent ($theGeom)
    {
        $theGeom = substr($theGeom, 4, strlen($theGeom));
        $theGeom = substr($theGeom, 0, - 1);
        
        $theGeom = explode(',', $theGeom);
        $theGeom[0] = explode(' ', $theGeom[0]);
        $theGeom[1] = explode(' ', $theGeom[1]);
        
        $extent = array();
        $extent['min']['x'] = floatval($theGeom[0][0]);
        $extent['min']['y'] = floatval($theGeom[0][1]);
        $extent['max']['x'] = floatval($theGeom[1][0]);
        $extent['max']['y'] = floatval($theGeom[1][1]);
        
        return $extent;
    }
    
    /**
     * Convert bbox wkt representation into an array
     * 
     * @param string $geometry Geometry WKT string
     * @return array 
     */
    public static function extentToWktPolygon ($extent)
    {
        $minx = $extent['min']['x'];
        $miny = $extent['min']['y'];
        $maxx = $extent['max']['x'];
        $maxy = $extent['max']['y'];
        
        $polyStr = "POLYGON(({$minx} {$miny}, {$maxx} {$miny}, {$maxx} {$maxy}, {$minx} {$maxy}, {$minx} {$miny}))";
        
        return $polyStr;
    }

    /**
     * Get the list of available fonts
     * 
     * @param $options <ul>Available options:
     * <li>forForm bool - Data will be returned as an array suitable for setMultiOptions method of Zend_Form</li></ul>
     * @return array
     */
    public static function getFontList (array $options = array())
    {
        $result = $fonts = array();
        
        //Write
        $contents = '';
        $paths = array(
            GT_CONFIG_PATH . DIR_SEP . 'fonts' . DIR_SEP . '*.ttf' ,
            GT_CONFIG_PATH . DIR_SEP . 'fonts' . DIR_SEP . '*.TTF' ,
                          
            //Ubuntu
            '/usr/share/fonts/truetype/freefont/*.ttf' ,
            '/usr/share/fonts/truetype/ttf-liberation/*.ttf' ,
            
            //CentOS, RHEL
            '/usr/share/fonts/freefont/*.ttf' ,
            '/usr/share/fonts/gnu-free/*.ttf' ,
            '/usr/share/fonts/liberation/*.ttf' ,
            
            //openSUSE, SLES
            '/usr/share/fonts/truetype/Free*.ttf' ,
            '/usr/share/fonts/truetype/Liberation*.ttf' ,
             
            //Microsoft Windows
            "C:\\WINDOWS\\Fonts\\arial*.ttf" ,
            "C:\\WINDOWS\\Fonts\\cour*.ttf" ,
            "C:\\WINDOWS\\Fonts\\times*.ttf" ,
            "C:\\WINDOWS\\Fonts\\verdana*.ttf" ,
        );
        foreach($paths as $path) {
            $fonts = array_merge($fonts, glob($path));    
        }
        foreach ($fonts as $key => $value) {
            $pathinfo = pathinfo($value);
            $contents .= $pathinfo['filename'] . ' ' . $value . PHP_EOL;
        }
        file_put_contents(GT_CONFIG_PATH . DIR_SEP . 'fonts' . DIR_SEP . 'fonts.txt', $contents);

        //Read
        $contents = file(GT_CONFIG_PATH . DIR_SEP . 'fonts' . DIR_SEP . 'fonts.txt');
        foreach ($contents as $line) {
            $line = explode(' ', $line);
            if (isset($options['forForm']) && $options['forForm'] = true) {
                $result[$line[0]] = $line[0];
            } else {
                $result[] = $line[0];
            }
        }

        return $result;
    
    }

    /**
     * Get the list of available symbols
     * 
     * @param $options <ul>Available options:
     * <li>forForm bool - Data will be returned as an array suitable for setMultiOptions method of Zend_Form</li></ul>
     * @return array
     */
    public static function getSymbolList (array $options = array())
    {
        $contents = file_get_contents(GT_CONFIG_PATH . DIR_SEP . 'symbols.sym');
        
        preg_match_all("/name\s+[\'\"](.*?)[\'\"]/i", $contents, $matches);
        
        if (isset($options['forForm']) && $options['forForm'] = true) {
            $result = array(
                '' => ''
            );
            foreach ($matches[1] as $symbol) {
                $result[$symbol] = $symbol;
            }
            return $result;
        } else {
            return $matches[1];
        }
    
    }

    /**
     * How many inches 1 $unit has
     * 
     * @param string $unit Unit name (inches, ft, mi, m, km, dd, yd)
     * @return float
     */
    public static function inchesPerUnit ($unit)
    {
        
        $inchesPerUnit = array(
            'inches' => 1.0 , 
            'ft' => 12.0 , 
            'mi' => 63360.0 , 
            'm' => 39.3701 , 
            'km' => 39370.1 , 
            'dd' => 4374754 , 
            'yd' => 36
        );
        
        if (! in_array($unit, array_keys($inchesPerUnit))) {
            throw new Giscuit_Exception('Invalid map units');
        }
        
        return $inchesPerUnit[$unit];
    }

    /**
     * Conver unit name to MapScript constant
     * 
     * @param string $unit Unit name (inches, ft, mi, m, km, dd)
     * @return float
     */
    public static function unitsToValue ($unit)
    {
        if ($unit == 'm') {
            return MS_METERS;
        } elseif ($unit == 'dd') {
            return MS_DD;
        } elseif ($unit == 'inches') {
            return MS_INCHES;
        } elseif ($unit == 'ft') {
            return MS_FEET;
        } elseif ($unit == 'mi') {
            return MS_MILES;
        } elseif ($unit == 'km') {
            return MS_KILOMETERS;
        } else {
            throw new Giscuit_Exception('Invalid map units');
        }
    }

    /**
     * Convert scale to resolution
     * 
     * @param float $resolution Resolution
     * @param integer $dpi Dpi
     * @return float
     */
    public static function getScaleFromResolution ($resolution, $dpi = self::RESOLUTION)
    {
        $config = Zend_Registry::get('config');
        
        $scale = $resolution * self::inchesPerUnit($config->map->units) * $dpi;
        
        return $scale;
    }

    /**
     * Get publish map options
     * 
     * @return array
     */
    public static function getPublishOptions ()
    {
        return array(
            'addEditLayer' => true , 
            'orderDirection' => 'desc' , 
            'wherePublish' => 't' , 
            'whereActive' => 't' , 
            'withFields' => true , 
            'withClassesAndStyles' => true , 
            'withPublish' => true ,
            'withCluster' => true
        );
    }

    /**
     * Get default map options
     * 
     * @return array
     */
    public static function getDefaultOptions ()
    {
        return array(
            'addEditLayer' => false , 
            'orderDirection' => 'desc' , 
            'whereView' => 't' , 
            'whereActive' => 't' , 
            'withLayerSettings' => true , 
            'withAcl' => true , 
            'withFields' => true , 
            'withClassesAndStyles' => true ,
            'withCluster' => true
        );
    }

    /**
     * Generate a raster image using user's current layer settings
     * 
     * @param string $file File name
     * @param $options <ul>Available options:
     * <li>mapOptions array - Map paramaters</li> 
     * <li>resolution integer - Resolution</li>
     * <li>extent array - Extent</li>
     * <li>scale float - Scale</li>
     * <li>burnPolygon string - Burn area outside the polygon</li>
     * <li>watermark bool - Watermark the image</li></ul>
     *  
     * @return boolean
     */
    public static function getImage ($file, array $options = array())
    {
        $config = Zend_Registry::get('config');
        set_time_limit(0);
        
        if (empty($options['mapOptions'])) {
            $mapOptions = Giscuit_Map_Map::getDefaultOptions();
            $mapOptions['addEditLayer'] = true;
            $options['mapOptions'] = $mapOptions;
        }
                
        //Create map
        $map = new Giscuit_Map_Map(GT_CONFIG_PATH . DIR_SEP . 'default.map');
        $map->selectOutputFormat('jpeg');
        
        //Zend_Registry::set('map', $map);
        
        $map->setParamsFromConfig();
                
        //Set resolution
        $map->set('resolution', $options['resolution']);
        
        //Set size
        $info = Giscuit_Map_Map::getInfoFromExtent($options['extent'], $options['scale'], $options['resolution']);
        
        //Create world file
        $wdl = Giscuit_Map_Map::extent2wld($info);
        
        $map->setExtent($info['extent']);
        $map->setSize($info['size']);
        $map->setPixelSize();
        
        
        if (! empty($options['watermark'])) {
            $options['mapOptions']['watermark'] = $options['watermark'];
        }
        
        $map->addCurrentLayers($options['mapOptions']);
        $map->selectOutputFormat('jpeg');
        
        $fullImagePath = GT_TMP_PATH . DIR_SEP . $file . ".jpg";
        
        //Create image
        $image = $map->draw();
        $image->saveimage($fullImagePath);
        
        unset($map);
        unset($image);
        
        //Create world file
        file_put_contents(GT_TMP_PATH . DIR_SEP . $file . ".jgw", $wdl);
        if (! empty($options['burnPolygon'])) {
            
            //Convert to tiff
            Giscuit_Exec::exec('gdal_translate', "-of GTiff -co TILED=YES -mo TIFFTAG_XRESOLUTION={$options['resolution']} -mo TIFFTAG_YRESOLUTION={$options['resolution']} -co COMPRESS=LZW -a_srs EPSG:{$config->map->srid} \"" . GT_TMP_PATH . DIR_SEP . "$file.jpg\" \"" . GT_TMP_PATH . DIR_SEP . "{$file}_burn.tif\"", GT_GDAL_BIN_PATH);
            unlink(GT_TMP_PATH . DIR_SEP . "{$file}.jpg");
            
            $connectionString = Giscuit_Controller_Plugin_Init::getDbConnectionString();
            $connectionString .= ' tables=' . $config->database->schema . "_layers.rastercut";
            
            //Burn polygon
            Giscuit_Exec::exec('gdal_rasterize', "-i -b 1 -b 2 -b 3 -burn 255 -burn 255 -burn 255 PG:\"$connectionString\" -sql \"SELECT {$options['burnPolygon']} as the_geom\" \"" . GT_TMP_PATH . DIR_SEP . "{$file}_burn.tif\"", GT_GDAL_BIN_PATH);
            
            //Convert to jpeg
            Giscuit_Exec::exec('gdal_translate', "-of JPEG \"" . GT_TMP_PATH . DIR_SEP . "{$file}_burn.tif\" \"" . GT_TMP_PATH . DIR_SEP . "{$file}_out.jpg\"", GT_GDAL_BIN_PATH);
            
            unlink(GT_TMP_PATH . DIR_SEP . "{$file}_burn.tif");
            
            if (file_exists(GT_TMP_PATH . DIR_SEP . "{$file}_out.jpg.aux.xml")) {
                unlink(GT_TMP_PATH . DIR_SEP . "{$file}_out.jpg.aux.xml");
            }
            
            rename(GT_TMP_PATH . DIR_SEP . "{$file}_out.jpg", GT_TMP_PATH . DIR_SEP . "{$file}.jpg");
        }
        
        return true;
    }

    /**
     * Create a world file from extent
     * 
     * @param array $info $info['extent'] to get the extent
     * @return string
     */
    public static function extent2wld ($info)
    {
        $coef = 0.5;
        
        $pxW = ($info['extent']['max']['x'] - $info['extent']['min']['x']) / $info['size']['width']; //A
        $pxH = ($info['extent']['max']['y'] - $info['extent']['min']['y']) / $info['size']['height'] * - 1; //E
        

        $topLeftX = $info['extent']['min']['x'] + $pxW * $coef; //C
        $topLeftY = $info['extent']['max']['y'] + $pxH * $coef; //F
        

        $pxW = number_format($pxW, 12, '.', '');
        $pxH = number_format($pxH, 12, '.', '');
        
        return "{$pxW}\n0\n0\n{$pxH}\n{$topLeftX}\n{$topLeftY}";
    }

    /**
     * Using extent, scale and resolution calculate image and pixel size
     * 
     * @param $extent Extent
     * @param $scale Scale
     * @param $resolution Resolution
     * @return array
     */
    public static function getInfoFromExtent ($extent, $scale, $resolution)
    {
        
        $config = Zend_Registry::get('config');
        
        $factor = Giscuit_Map_Map::inchesPerUnit($config->map->units);
        
        $sizeW = ($resolution * $factor * ($extent['max']['x'] - $extent['min']['x'])) / $scale;
        $sizeH = ($resolution * $factor * ($extent['max']['y'] - $extent['min']['y'])) / $scale;
        
        $pxW = ($extent['max']['x'] - $extent['min']['x']) / $sizeW;
        $pxH = ($extent['max']['y'] - $extent['min']['y']) / $sizeH;
        
        $sizeW = floor($sizeW);
        $sizeH = floor($sizeH);
        
        $extent['max']['x'] = $extent['min']['x'] + $sizeW * $pxW;
        $extent['max']['y'] = $extent['min']['y'] + $sizeH * $pxH;
        
        return array(
            'size' => array(
                'width' => $sizeW , 
                'height' => $sizeH
            ) , 
            'pixel' => array(
                'width' => $pxW , 
                'height' => $pxH
            ) , 
            'extent' => array(
                'min' => array(
                    'x' => $extent['min']['x'] , 
                    'y' => $extent['min']['y']
                ) , 
                'max' => array(
                    'x' => $extent['max']['x'] , 
                    'y' => $extent['max']['y']
                )
            )
        );
    }

    /**
     * Create an error image and send it's output to output buffers
     * 
     * @return void
     */
    public static function drawError ($text)
    {
        if (! empty($_GET['map_size'])) {
            $mapSize = $_GET['map_size'];
        } elseif (! empty($_GET['MAP_SIZE'])) {
            $mapSize = $_GET['MAP_SIZE'];
        }
        if (! empty($mapSize)) {
            $mapSize = explode(' ', $_GET['map_size']);
            if (count($mapSize) == 2) {
                $width = floatval($mapSize[0]);
                $height = floatval($mapSize[1]);
            }
        }
        
        // Set font size
        $font_size = 4;
        
        //Offset
        $offSet = 150;
        
        $ts = explode("\n", $text);
        
        $el = imagefontheight($font_size);
        $em = imagefontwidth($font_size);
        
        //Create the image pallette
        $img = imagecreatetruecolor($width, $height);
        
        //Dark red background
        $bg = imagecolorallocate($img, 255, 255, 255);
        imagefilledrectangle($img, 0, 0, $width, $height, $bg);
        
        //White font color
        $color = imagecolorallocate($img, 0, 0, 0);
        
        foreach ($ts as $k => $string) {
            //Length of the string
            $len = strlen($string);
            
            //Y-coordinate of character, X changes, Y is static
            $ypos = 0;
            
            //Loop through the string
            for ($i = 0; $i < $len; $i ++) {
                                
                //Position of the character horizontally
                $xpos = $offSet + $i * $em;
                $ypos = $offSet + $k * $el;
                
                //Draw character
                imagechar($img, $font_size, $xpos, $ypos, $string, $color);
                
                //Remove character from string
                $string = substr($string, 1);
            }
        }
        
        //Return the image
        header("Content-Type: image/png");
        imagepng($img);
        
        //Remove image
        imagedestroy($img);
    }
    
    /**
     * Clear legend cache
     *  
     * @return void
     */
    public static function clearLegend()
    {
        $files = glob(GT_PUBLIC_PATH . DIR_SEP . 'legend' . DIR_SEP . '*.png');
        foreach ($files as $file) {
            if (is_writeable($file)) {
                unlink($file);
            }
        }
    }
    
    /**
     * Clear link cache
     *  
     * @return void
     */
    public static function clearLink()
    {
        $files = glob(GT_PUBLIC_PATH . DIR_SEP . 'link' . DIR_SEP . '*.html');
        foreach ($files as $file) {
            if (is_writeable($file)) {
                unlink($file);
            }
        }
    }
    
    /**
     * Returns a layer from the map given an index
     * 
     * @param integer $index Index value (0 = first layer)
     * @return Giscuit_Map_Layer_Abstract
     */
     public function getLayer($index) {
        $name = $this->getMsObj()->getLayer($index)->name;    
        
        if(isset($this->children[$name])) {
            return $this->children[$name];
        } else {
            throw new Giscuit_Exception("Layer ({$name}) was not found in the haystack.");
        }
    }
    
    /**
     * Returns a layer from the map given a layer name.
     * 
     * @param string $name Layer name
     * @return Giscuit_Map_Layer_Abstract
     */
    public function getLayerByName($name) {
        if(isset($this->children[$name])) {
            return $this->children[$name];
        } else {
            throw new Giscuit_Exception("Layer ({$name}) was not found in the haystack.");
        }
    }
    
    /**
     * Remove a layer from the map
     * 
     * @param integer $index Index value (0 = first layer)
     * @return Giscuit_Map_Layer_Abstract Returns the removed layer on success, else null.
     */
    public function removeLayer($index) {
        $name = $this->getLayer($index)->name;
        if(isset($this->children[$name])) {
            $return = $this->getMsObj()->removeLayer($index);
            unset($this->children[$name]);

            return $return;
        } else {
            throw new Giscuit_Exception("Layer ({$name}) was not found in the haystack.");    
        }
        
    }
    
    public function insertLayer($layer, $index = -1)
    {
        throw new Giscuit_Exception("Method not implemented");
    }
    
    /**
     * Get the list of available output formats
     * 
     * @param $options <ul>Available options:</ul>
     * @return array
     */
    public static function getOutputformatsList (array $options = array())
    {
        $exclude = array(
            "png",
              "png24",
              "jpeg",
        );
        
        $contents = file_get_contents(GT_CONFIG_PATH . DIR_SEP . 'default.map');
        
        preg_match_all("/outputformat(.*?)name\s+[\'\"](.*?)[\'\"](.*?)end/is", $contents, $matches);
        
        $result = array_values(array_diff($matches[2], $exclude));
        
        return $result;
    }
    
}
